==> magento-module-saas-order-sync-0.1.0.zip: Test/Unit/Core/OrderSync/OrderSyncManagerTest.php <==
<?php

declare(strict_types=1);

namespace Magento\SaaSOrderSync\Test\Unit\Core\OrderSync;

use DateTime;
use DateTimeInterface;
use Exception;
use Magento\Framework\Serialize\Serializer\Json;
use Magento\Framework\Webapi\Exception as WebapiException;
use Magento\SaaSOrderSync\Core\Bulk\BulkAdapter;
use Magento\SaaSOrderSync\Core\CommerceDataExport\OrdersDataExporterAdapter;
use Magento\SaaSOrderSync\Core\Common\Result;
use Magento\SaaSOrderSync\Core\Merchant\MerchantSaaSClient;
use Magento\SaaSOrderSync\Core\OrderSync\LocalOrderSyncRepository;
use Magento\SaaSOrderSync\Core\OrderSync\OrdersBulkManager;
use Magento\SaaSOrderSync\Core\OrderSync\OrderSyncLock;
use Magento\SaaSOrderSync\Core\OrderSync\OrderSyncSaaSClient;
use Magento\SaaSOrderSync\Core\OrderSync\OrderSyncManager;
use Magento\SaaSOrderSync\Test\Unit\Core\Common\Generators;
use Magento\SaaSOrderSync\Test\Unit\Core\Common\Time\ClockMock;
use Magento\SaaSOrderSync\Test\Unit\Core\SaaS\SaaSClientMockResolver;
use function PHPUnit\Framework\anything;
use function PHPUnit\Framework\arrayHasKey;
use function PHPUnit\Framework\assertEquals;
use function PHPUnit\Framework\assertNotNull;
use function PHPUnit\Framework\assertThat;
use function PHPUnit\Framework\callback;
use PHPUnit\Framework\Constraint\Constraint;
use PHPUnit\Framework\Constraint\LogicalAnd;
use function PHPUnit\Framework\equalTo;
use function PHPUnit\Framework\exactly;
use function PHPUnit\Framework\isInstanceOf;
use PHPUnit\Framework\MockObject\MockObject;
use function PHPUnit\Framework\logicalNot;
use function PHPUnit\Framework\never;
use function PHPUnit\Framework\once;
use PHPUnit\Framework\TestCase;
use Psr\Log\NullLogger;

class OrderSyncManagerTest extends TestCase
{
    private OrderSyncSaaSClient&MockObject $orderSyncClient;
    private OrdersDataExporterAdapter&MockObject $exporterAdapterMock;
    private MockObject&MerchantSaaSClient $merchantClient;
    private BulkAdapter&MockObject $bulkAdapterMock;
    private LocalOrderSyncRepository&MockObject $localOrderSyncRepositoryMock;
    private ClockMock $clockMock;
    private FixedIdentityGenerator $identityGenerator;

    private OrderSyncManager $orderSyncManager;

    public function setUp(): void
    {
        $logger = new NullLogger();

        $this->identityGenerator = new FixedIdentityGenerator();

        $this->exporterAdapterMock = $this->createMock(OrdersDataExporterAdapter::class);
        $this->bulkAdapterMock = $this->createMock(BulkAdapter::class);
        $this->localOrderSyncRepositoryMock = $this->createMock(LocalOrderSyncRepository::class);

        $clientResolver = new SaaSClientMockResolver($this);
        $this->orderSyncClient = $clientResolver->createOrderSyncClient();
        $this->merchantClient = $clientResolver->createMerchantClient();

        $this->clockMock = new ClockMock();

        $this->orderSyncManager = new OrderSyncManager(
            $this->identityGenerator,
            $this->exporterAdapterMock,
            new OrdersBulkManager(
                $this->exporterAdapterMock,
                new Json(),
                $this->localOrderSyncRepositoryMock,
                $logger,
                $this->bulkAdapterMock
            ),
            $clientResolver,
            $logger,
            new OrderSyncLock($logger, new NoLock()),
            $this->clockMock
        );
    }

    public function test_shouldCreateSyncUntilCurrentTime_whenThereAreOrders()
    {
        $syncId = $this->identityGenerator->getId();
        $createdFrom = (new DateTime('2022-01-01'))->format(DateTimeInterface::ATOM);
        $createdFromDateTime = new DateTime($createdFrom);

        $this->clockMock->setNow('2023-01-01');
        $createdToDateTime = $this->clockMock->now();
        $clientCode = 'unit-test';

        $this->merchantClient->expects(once())
            ->method('getDataSummary')
            ->willReturn(['orders' => []]);

        $this->exporterAdapterMock->expects(once())
            ->method('count')
            ->with($createdFromDateTime, $createdToDateTime)
            ->willReturn(5);

        $this->exporterAdapterMock->expects(once())
            ->method('findOrderIds')
            ->with($createdFromDateTime, $createdToDateTime)
            ->willReturn(Generators::fixed([1, 2], [3, 4], [5]));

        $this->orderSyncClient->expects(once())
            ->method('createOrderSync')
            ->with($syncId, $createdFromDateTime, $createdToDateTime, 5, $clientCode)
            ->willReturnCallback($this->orderSyncResponse());

        $bulkArg = $this->and(
            $this->arrayHasKeyAndValue('uuid', $syncId),
            arrayHasKey('description')
        );
        $this->bulkAdapterMock->expects(exactly(3))
            ->method('scheduleOperation')
            ->withConsecutive(
                [$bulkArg, 1, ['orderIds' => [1, 2]]],
                [$bulkArg, 2, ['orderIds' => [3, 4]]],
                [$bulkArg, 3, ['orderIds' => [5]]],
            )
            ->willReturn(Result::data());

        $response = $this->orderSyncManager->createOrderSync($createdFrom, $clientCode);

        assertNotNull($response->getOrderSync()->getSyncId());
    }

    public function test_shouldCreateSyncWithOptimizedTimeRange_whenThereAreOrders()
    {
        $syncId = $this->identityGenerator->getId();
        $createdFrom = (new DateTime('2022-01-01'))->format(DateTimeInterface::ATOM);
        $createdFromDateTime = new DateTime($createdFrom);
        $clientCode = 'unit-test';

        $dataSummaryCreatedFrom = (new DateTime('2022-06-01'))->format(DateTimeInterface::ATOM);
        $dataSummaryCreatedFromDateTime = new DateTime($dataSummaryCreatedFrom);

        $this->merchantClient->expects(once())
            ->method('getDataSummary')
            ->willReturn(['orders' => ['createdFrom' => $dataSummaryCreatedFrom]]);

        $this->exporterAdapterMock->expects(once())
            ->method('count')
            ->with($createdFromDateTime, $dataSummaryCreatedFromDateTime)
            ->willReturn(1);

        $this->exporterAdapterMock->expects(once())
            ->method('findOrderIds')
            ->with($createdFromDateTime, $dataSummaryCreatedFromDateTime)
            ->willReturn(Generators::fixed([1]));

        $this->orderSyncClient->expects(once())
            ->method('createOrderSync')
            ->with($syncId, $createdFromDateTime, $dataSummaryCreatedFromDateTime, 1, $clientCode)
            ->willReturnCallback($this->orderSyncResponse());

        $bulkArg = $this->and(
            $this->arrayHasKeyAndValue('uuid', $syncId),
            arrayHasKey('description')
        );
        $this->bulkAdapterMock->expects(once())
            ->method('scheduleOperation')
            ->with($bulkArg, 1, ['orderIds' => [1]])
            ->willReturn(Result::data());

        $response = $this->orderSyncManager->createOrderSync($createdFrom, $clientCode);

        assertNotNull($response->getOrderSync()->getSyncId());
    }

    public function test_shouldCreateSyncUntilCurrentTime_whenOptimizedTimeRangeIsOff()
    {
        $syncId = $this->identityGenerator->getId();
        $createdFrom = (new DateTime('2022-01-01'))->format(DateTimeInterface::ATOM);
        $createdFromDateTime = new DateTime($createdFrom);
        $clientCode = 'unit-test';

        $this->clockMock->setNow('2023-01-01');
        $createdToDateTime = $this->clockMock->now();

        $this->merchantClient->expects(never())
            ->method(anything());

        $this->exporterAdapterMock->expects(once())
            ->method('count')
            ->with($createdFromDateTime, $createdToDateTime)
            ->willReturn(1);

        $this->exporterAdapterMock->expects(once())
            ->method('findOrderIds')
            ->with($createdFromDateTime, $createdToDateTime)
            ->willReturn(Generators::fixed([1]));

        $this->orderSyncClient->expects(once())
            ->method('createOrderSync')
            ->with($syncId, $createdFromDateTime, $createdToDateTime, 1, $clientCode)
            ->willReturnCallback($this->orderSyncResponse());

        $bulkArg = $this->and(
            $this->arrayHasKeyAndValue('uuid', $syncId),
            arrayHasKey('description')
        );
        $this->bulkAdapterMock->expects(once())
            ->method('scheduleOperation')
            ->with($bulkArg, 1, ['orderIds' => [1]])
            ->willReturn(Result::data());

        $response = $this->orderSyncManager->createOrderSync($createdFrom, $clientCode, false);

        assertNotNull($response->getOrderSync()->getSyncId());
    }

    public function test_shouldCreateAndCompleteSync_whenTimeRangeIsAlreadySync()
    {
        $syncId = $this->identityGenerator->getId();
        $createdFrom = (new DateTime('2022-01-01'))->format(DateTimeInterface::ATOM);
        $createdFromDateTime = new DateTime($createdFrom);
        $clientCode = 'unit-test';

        $this->clockMock->setNow('2023-01-01');
        $createdToDateTime = $this->clockMock->now();
        $createdTo = $createdToDateTime->format(DateTimeInterface::ATOM);

        $dataSummaryCreatedFrom = (new DateTime('2021-01-01'))->format(DateTimeInterface::ATOM);

        $this->merchantClient->expects(once())
            ->method('getDataSummary')
            ->willReturn(['orders' => ['createdFrom' => $dataSummaryCreatedFrom]]);

        $this->exporterAdapterMock->expects(never())
            ->method(anything());

        $this->orderSyncClient->expects(once())
            ->method('createOrderSync')
            ->with($syncId, $createdFromDateTime, $createdToDateTime, 0, $clientCode)
            ->willReturnCallback($this->orderSyncResponse());

        $this->orderSyncClient->expects(once())
            ->method('completeOrderSync')
            ->with($syncId)
            ->willReturn(Result::data([
                'syncId' => $syncId,
                'createdFrom' => $createdFrom,
                'createdTo' => $createdTo,
                'initialCount' => 0,
                'clientCode' => $clientCode,
                'processedCount' => 0,
                'errorCount' => 0,
                'status' => 'COMPLETED',
            ]));

        $response = $this->orderSyncManager->createOrderSync($createdFrom, $clientCode);

        assertNotNull($response->getOrderSync()->getSyncId());
        assertEquals($createdFrom, $response->getOrderSync()->getCreatedFrom());
        assertEquals($createdTo, $response->getOrderSync()->getCreatedTo());
        assertEquals('COMPLETED', $response->getOrderSync()->getStatus());
        assertEquals(0, $response->getOrderSync()->getInitialCount());
        assertEquals(0, $response->getOrderSync()->getProcessedCount());
        assertEquals(0, $response->getOrderSync()->getErrorCount());
    }

    public function test_shouldCreateAndCompleteSync_whenNoOrdersInTheTimeRange()
    {
        $syncId = $this->identityGenerator->getId();
        $createdFrom = (new DateTime('2022-01-01'))->format(DateTimeInterface::ATOM);
        $createdFromDateTime = new DateTime($createdFrom);
        $clientCode = 'unit-test';

        $this->clockMock->setNow('2023-01-01');
        $createdToDateTime = $this->clockMock->now();
        $createdTo = $createdToDateTime->format(DateTimeInterface::ATOM);

        $this->merchantClient->expects(once())
            ->method('getDataSummary')
            ->willReturn(['orders' => []]);

        $this->exporterAdapterMock->expects(once())
            ->method('count')
            ->with($createdFromDateTime, $createdToDateTime)
            ->willReturn(0);

        $this->exporterAdapterMock->expects(never())
            ->method(logicalNot(equalTo('count')));

        $this->orderSyncClient->expects(once())
            ->method('createOrderSync')
            ->with($syncId, $createdFromDateTime, $createdToDateTime, 0, $clientCode)
            ->willReturnCallback($this->orderSyncResponse());

        $this->orderSyncClient->expects(once())
            ->method('completeOrderSync')
            ->with($syncId)
            ->willReturn(Result::data([
                'syncId' => $syncId,
                'createdFrom' => $createdFrom,
                'createdTo' => $createdTo,
                'initialCount' => 0,
                'clientCode' => $clientCode,
                'processedCount' => 0,
                'errorCount' => 0,
                'status' => 'COMPLETED',
            ]));

        $response = $this->orderSyncManager->createOrderSync($createdFrom, $clientCode);

        assertNotNull($response->getOrderSync()->getSyncId());
        assertEquals($createdFrom, $response->getOrderSync()->getCreatedFrom());
        assertEquals($createdTo, $response->getOrderSync()->getCreatedTo());
        assertEquals('COMPLETED', $response->getOrderSync()->getStatus());
        assertEquals(0, $response->getOrderSync()->getInitialCount());
        assertEquals(0, $response->getOrderSync()->getProcessedCount());
        assertEquals(0, $response->getOrderSync()->getErrorCount());
    }

    public function test_shouldFailWithBadRequest_whenCreatedFromIsNotValid()
    {
        $createdFrom = 'not-valid';
        $clientCode = 'unit-test';

        $this->assertThrowsWebapiException(
            fn () => $this->orderSyncManager->createOrderSync($createdFrom, $clientCode),
            WebapiException::HTTP_BAD_REQUEST
        );
    }

    public function test_shouldFailWithBadRequest_whenCreatedFromIsNotInThePast()
    {
        $this->clockMock->setNow('2023-01-01');
        $createdFrom = '2023-01-02';
        $clientCode = 'unit-test';

        $this->assertThrowsWebapiException(
            fn () => $this->orderSyncManager->createOrderSync($createdFrom, $clientCode),
            WebapiException::HTTP_BAD_REQUEST
        );
    }

    public function test_shouldFailWithBadRequest_whenClientIsNotValid()
    {
        $createdFrom = (new DateTime())->format(DateTimeInterface::ATOM);
        $clientCode = '';

        $this->assertThrowsWebapiException(
            fn () => $this->orderSyncManager->createOrderSync($createdFrom, $clientCode),
            WebapiException::HTTP_BAD_REQUEST
        );
    }

    private function arrayHasKeyAndValue(string $key, mixed $value): Constraint
    {
        return callback(fn (array $subject): bool => array_key_exists($key, $subject) && $subject[$key] == $value);
    }

    private function and(Constraint ...$constraints): Constraint
    {
        return LogicalAnd::fromConstraints(...$constraints);
    }

    private function assertThrowsWebapiException(callable $callable, int $httpStatusCode)
    {
        $this->assertThrowsException(
            $callable,
            WebapiException::class,
            fn (WebapiException $e) => assertThat($e->getHttpCode(), equalTo($httpStatusCode))
        );
    }

    private function assertThrowsException(callable $callable, string $exception, callable $exceptionCallable = null): void
    {
        try {
            $callable();
        } catch (Exception $e) {
            assertThat($e, isInstanceOf($exception));
            if ($exceptionCallable != null) {
                $exceptionCallable($e);
            }
        }
    }

    private function orderSyncResponse(string $status = 'IN_PROGRESS'): callable
    {
        return fn (string $syncId, DateTime $createdFrom, DateTime $createdTo, int $initialCount, string $clientCode): array => [
            'syncId' => $syncId,
            'createdFrom' => $createdFrom->format(DateTimeInterface::ATOM),
            'createdTo' => $createdTo->format(DateTimeInterface::ATOM),
            'initialCount' => $initialCount,
            'clientCode' => $clientCode,
            'processedCount' => 0,
            'errorCount' => 0,
            'status' => $status,
        ];
    }

}

==> magento-module-saas-order-sync-0.2.0.zip: Test/Unit/Core/OrderSync/OrderSyncManagerTest.php <==
<?php

declare(strict_types=1);

namespace Magento\SaaSOrderSync\Test\Unit\Core\OrderSync;

use DateTime;
use DateTimeInterface;
use Exception;
use Magento\Framework\Serialize\Serializer\Json;
use Magento\Framework\Webapi\Exception as WebapiException;
use Magento\SaaSOrderSync\Api\OrderSync\Bulk\OperationData;
use Magento\SaaSOrderSync\Api\Result;
use Magento\SaaSOrderSync\Core\Bulk\BulkAdapter;
use Magento\SaaSOrderSync\Core\CommerceDataExport\OrdersDataExporterAdapter;
use Magento\SaaSOrderSync\Core\Merchant\MerchantSaaSClient;
use Magento\SaaSOrderSync\Core\OrderSync\Bulk\Manager as OrderSyncBulkManager;
use Magento\SaaSOrderSync\Core\OrderSync\LocalOrderSyncRepository;
use Magento\SaaSOrderSync\Core\OrderSync\Lock;
use Magento\SaaSOrderSync\Core\OrderSync\Manager as OrderSyncManager;
use Magento\SaaSOrderSync\Core\OrderSync\SaaSClient;
use Magento\SaaSOrderSync\Test\Unit\Core\Common\Generators;
use Magento\SaaSOrderSync\Test\Unit\Core\Common\Time\ClockMock;
use Magento\SaaSOrderSync\Test\Unit\Core\SaaS\SaaSClientMockResolver;
use function PHPUnit\Framework\anything;
use function PHPUnit\Framework\arrayHasKey;
use function PHPUnit\Framework\assertEquals;
use function PHPUnit\Framework\assertNotNull;
use function PHPUnit\Framework\assertThat;
use function PHPUnit\Framework\callback;
use PHPUnit\Framework\Constraint\Constraint;
use PHPUnit\Framework\Constraint\LogicalAnd;
use function PHPUnit\Framework\equalTo;
use function PHPUnit\Framework\exactly;
use function PHPUnit\Framework\isInstanceOf;
use function PHPUnit\Framework\logicalNot;
use PHPUnit\Framework\MockObject\MockObject;
use function PHPUnit\Framework\never;
use function PHPUnit\Framework\once;
use PHPUnit\Framework\TestCase;
use Psr\Log\NullLogger;

class OrderSyncManagerTest extends TestCase
{
    private SaaSClient&MockObject $orderSyncClient;
    private OrdersDataExporterAdapter&MockObject $exporterAdapterMock;
    private MockObject&MerchantSaaSClient $merchantClient;
    private BulkAdapter&MockObject $bulkAdapterMock;
    private LocalOrderSyncRepository&MockObject $localOrderSyncRepositoryMock;
    private ClockMock $clockMock;
    private FixedIdentityGenerator $identityGenerator;

    private OrderSyncManager $orderSyncManager;

    public function setUp(): void
    {
        $logger = new NullLogger();

        $this->identityGenerator = new FixedIdentityGenerator();

        $this->exporterAdapterMock = $this->createMock(OrdersDataExporterAdapter::class);
        $this->bulkAdapterMock = $this->createMock(BulkAdapter::class);
        $this->localOrderSyncRepositoryMock = $this->createMock(LocalOrderSyncRepository::class);

        $clientResolver = new SaaSClientMockResolver($this);
        $this->orderSyncClient = $clientResolver->createOrderSyncClient();
        $this->merchantClient = $clientResolver->createMerchantClient();

        $this->clockMock = new ClockMock();

        $this->orderSyncManager = new OrderSyncManager(
            $this->identityGenerator,
            $this->exporterAdapterMock,
            new OrderSyncBulkManager(
                $this->exporterAdapterMock,
                new Json(),
                $this->localOrderSyncRepositoryMock,
                $logger,
                $this->bulkAdapterMock
            ),
            $clientResolver,
            $logger,
            new Lock($logger, new NoLock()),
            $this->clockMock
        );

        $this->bulkAdapterMock->method('isAMQPAvailable')
            ->withAnyParameters()
            ->willReturn(true);
    }

    public function test_shouldCreateSyncUntilCurrentTime_whenThereAreOrders()
    {
        $syncId = $this->identityGenerator->getId();
        $createdFrom = (new DateTime('2022-01-01'))->format(DateTimeInterface::ATOM);
        $createdFromDateTime = new DateTime($createdFrom);

        $this->clockMock->setNow('2023-01-01');
        $createdToDateTime = $this->clockMock->now();
        $clientCode = 'unit-test';

        $this->merchantClient->expects(once())
            ->method('getDataSummary')
            ->willReturn(['orders' => []]);

        $this->exporterAdapterMock->expects(once())
            ->method('count')
            ->with($createdFromDateTime, $createdToDateTime)
            ->willReturn(5);

        $this->exporterAdapterMock->expects(once())
            ->method('findOrderIds')
            ->with($createdFromDateTime, $createdToDateTime)
            ->willReturn(Generators::fixed([1, 2], [3, 4], [5]));

        $this->orderSyncClient->expects(once())
            ->method('createOrderSync')
            ->with($syncId, $createdFromDateTime, $createdToDateTime, 5, $clientCode)
            ->willReturnCallback($this->orderSyncResponse());

        $bulkArg = $this->and(
            $this->arrayHasKeyAndValue('uuid', $syncId),
            arrayHasKey('description')
        );
        $this->bulkAdapterMock->expects(exactly(3))
            ->method('scheduleOperation')
            ->withConsecutive(
                [$bulkArg, 1, new OperationData(['orderIds' => [1, 2]])],
                [$bulkArg, 2, new OperationData(['orderIds' => [3, 4]])],
                [$bulkArg, 3, new OperationData(['orderIds' => [5]])],
            )
            ->willReturn(Result::data());

        $response = $this->orderSyncManager->createOrderSync($createdFrom, $clientCode);

        assertNotNull($response->getOrderSync()->getSyncId());
    }

    public function test_shouldCreateSyncWithOptimizedTimeRange_whenThereAreOrders()
    {
        $syncId = $this->identityGenerator->getId();
        $createdFrom = (new DateTime('2022-01-01'))->format(DateTimeInterface::ATOM);
        $createdFromDateTime = new DateTime($createdFrom);
        $clientCode = 'unit-test';

        $dataSummaryCreatedFrom = (new DateTime('2022-06-01'))->format(DateTimeInterface::ATOM);
        $dataSummaryCreatedFromDateTime = new DateTime($dataSummaryCreatedFrom);

        $this->merchantClient->expects(once())
            ->method('getDataSummary')
            ->willReturn(['orders' => ['createdFrom' => $dataSummaryCreatedFrom]]);

        $this->exporterAdapterMock->expects(once())
            ->method('count')
            ->with($createdFromDateTime, $dataSummaryCreatedFromDateTime)
            ->willReturn(1);

        $this->exporterAdapterMock->expects(once())
            ->method('findOrderIds')
            ->with($createdFromDateTime, $dataSummaryCreatedFromDateTime)
            ->willReturn(Generators::fixed([1]));

        $this->orderSyncClient->expects(once())
            ->method('createOrderSync')
            ->with($syncId, $createdFromDateTime, $dataSummaryCreatedFromDateTime, 1, $clientCode)
            ->willReturnCallback($this->orderSyncResponse());

        $bulkArg = $this->and(
            $this->arrayHasKeyAndValue('uuid', $syncId),
            arrayHasKey('description')
        );
        $this->bulkAdapterMock->expects(once())
            ->method('scheduleOperation')
            ->with($bulkArg, 1, new OperationData(['orderIds' => [1]]))
            ->willReturn(Result::data());

        $response = $this->orderSyncManager->createOrderSync($createdFrom, $clientCode);

        assertNotNull($response->getOrderSync()->getSyncId());
    }

    public function test_shouldCreateSyncUntilCurrentTime_whenOptimizedTimeRangeIsOff()
    {
        $syncId = $this->identityGenerator->getId();
        $createdFrom = (new DateTime('2022-01-01'))->format(DateTimeInterface::ATOM);
        $createdFromDateTime = new DateTime($createdFrom);
        $clientCode = 'unit-test';

        $this->clockMock->setNow('2023-01-01');
        $createdToDateTime = $this->clockMock->now();

        $this->merchantClient->expects(never())
            ->method(anything());

        $this->exporterAdapterMock->expects(once())
            ->method('count')
            ->with($createdFromDateTime, $createdToDateTime)
            ->willReturn(1);

        $this->exporterAdapterMock->expects(once())
            ->method('findOrderIds')
            ->with($createdFromDateTime, $createdToDateTime)
            ->willReturn(Generators::fixed([1]));

        $this->orderSyncClient->expects(once())
            ->method('createOrderSync')
            ->with($syncId, $createdFromDateTime, $createdToDateTime, 1, $clientCode)
            ->willReturnCallback($this->orderSyncResponse());

        $bulkArg = $this->and(
            $this->arrayHasKeyAndValue('uuid', $syncId),
            arrayHasKey('description')
        );
        $this->bulkAdapterMock->expects(once())
            ->method('scheduleOperation')
            ->with($bulkArg, 1, new OperationData(['orderIds' => [1]]))
            ->willReturn(Result::data());

        $response = $this->orderSyncManager->createOrderSync($createdFrom, $clientCode, false);

        assertNotNull($response->getOrderSync()->getSyncId());
    }

    public function test_shouldCreateAndCompleteSync_whenTimeRangeIsAlreadySync()
    {
        $syncId = $this->identityGenerator->getId();
        $createdFrom = (new DateTime('2022-01-01'))->format(DateTimeInterface::ATOM);
        $createdFromDateTime = new DateTime($createdFrom);
        $clientCode = 'unit-test';

        $this->clockMock->setNow('2023-01-01');
        $createdToDateTime = $this->clockMock->now();
        $createdTo = $createdToDateTime->format(DateTimeInterface::ATOM);

        $dataSummaryCreatedFrom = (new DateTime('2021-01-01'))->format(DateTimeInterface::ATOM);

        $this->merchantClient->expects(once())
            ->method('getDataSummary')
            ->willReturn(['orders' => ['createdFrom' => $dataSummaryCreatedFrom]]);

        $this->exporterAdapterMock->expects(never())
            ->method(anything());

        $this->orderSyncClient->expects(once())
            ->method('createOrderSync')
            ->with($syncId, $createdFromDateTime, $createdToDateTime, 0, $clientCode)
            ->willReturnCallback($this->orderSyncResponse());

        $this->orderSyncClient->expects(once())
            ->method('completeOrderSync')
            ->with($syncId)
            ->willReturn(Result::data([
                'syncId' => $syncId,
                'createdFrom' => $createdFrom,
                'createdTo' => $createdTo,
                'initialCount' => 0,
                'clientCode' => $clientCode,
                'processedCount' => 0,
                'errorCount' => 0,
                'status' => 'COMPLETED',
            ]));

        $response = $this->orderSyncManager->createOrderSync($createdFrom, $clientCode);

        assertNotNull($response->getOrderSync()->getSyncId());
        assertEquals($createdFrom, $response->getOrderSync()->getCreatedFrom());
        assertEquals($createdTo, $response->getOrderSync()->getCreatedTo());
        assertEquals('COMPLETED', $response->getOrderSync()->getStatus());
        assertEquals(0, $response->getOrderSync()->getInitialCount());
        assertEquals(0, $response->getOrderSync()->getProcessedCount());
        assertEquals(0, $response->getOrderSync()->getErrorCount());
    }

    public function test_shouldCreateAndCompleteSync_whenNoOrdersInTheTimeRange()
    {
        $syncId = $this->identityGenerator->getId();
        $createdFrom = (new DateTime('2022-01-01'))->format(DateTimeInterface::ATOM);
        $createdFromDateTime = new DateTime($createdFrom);
        $clientCode = 'unit-test';

        $this->clockMock->setNow('2023-01-01');
        $createdToDateTime = $this->clockMock->now();
        $createdTo = $createdToDateTime->format(DateTimeInterface::ATOM);

        $this->merchantClient->expects(once())
            ->method('getDataSummary')
            ->willReturn(['orders' => []]);

        $this->exporterAdapterMock->expects(once())
            ->method('count')
            ->with($createdFromDateTime, $createdToDateTime)
            ->willReturn(0);

        $this->exporterAdapterMock->expects(never())
            ->method(logicalNot(equalTo('count')));

        $this->orderSyncClient->expects(once())
            ->method('createOrderSync')
            ->with($syncId, $createdFromDateTime, $createdToDateTime, 0, $clientCode)
            ->willReturnCallback($this->orderSyncResponse());

        $this->orderSyncClient->expects(once())
            ->method('completeOrderSync')
            ->with($syncId)
            ->willReturn(Result::data([
                'syncId' => $syncId,
                'createdFrom' => $createdFrom,
                'createdTo' => $createdTo,
                'initialCount' => 0,
                'clientCode' => $clientCode,
                'processedCount' => 0,
                'errorCount' => 0,
                'status' => 'COMPLETED',
            ]));

        $response = $this->orderSyncManager->createOrderSync($createdFrom, $clientCode);

        assertNotNull($response->getOrderSync()->getSyncId());
        assertEquals($createdFrom, $response->getOrderSync()->getCreatedFrom());
        assertEquals($createdTo, $response->getOrderSync()->getCreatedTo());
        assertEquals('COMPLETED', $response->getOrderSync()->getStatus());
        assertEquals(0, $response->getOrderSync()->getInitialCount());
        assertEquals(0, $response->getOrderSync()->getProcessedCount());
        assertEquals(0, $response->getOrderSync()->getErrorCount());
    }

    public function test_shouldFailWithBadRequest_whenCreatedFromIsNotValid()
    {
        $createdFrom = 'not-valid';
        $clientCode = 'unit-test';

        $this->assertThrowsWebapiException(
            fn () => $this->orderSyncManager->createOrderSync($createdFrom, $clientCode),
            WebapiException::HTTP_BAD_REQUEST
        );
    }

    public function test_shouldFailWithBadRequest_whenCreatedFromIsNotInThePast()
    {
        $this->clockMock->setNow('2023-01-01');
        $createdFrom = '2023-01-02';
        $clientCode = 'unit-test';

        $this->assertThrowsWebapiException(
            fn () => $this->orderSyncManager->createOrderSync($createdFrom, $clientCode),
            WebapiException::HTTP_BAD_REQUEST
        );
    }

    public function test_shouldFailWithBadRequest_whenClientIsNotValid()
    {
        $createdFrom = (new DateTime())->format(DateTimeInterface::ATOM);
        $clientCode = '';

        $this->assertThrowsWebapiException(
            fn () => $this->orderSyncManager->createOrderSync($createdFrom, $clientCode),
            WebapiException::HTTP_BAD_REQUEST
        );
    }

    private function arrayHasKeyAndValue(string $key, mixed $value): Constraint
    {
        return callback(fn (array $subject): bool => array_key_exists($key, $subject) && $subject[$key] == $value);
    }

    private function and(Constraint ...$constraints): Constraint
    {
        return LogicalAnd::fromConstraints(...$constraints);
    }

    private function assertThrowsWebapiException(callable $callable, int $httpStatusCode)
    {
        $this->assertThrowsException(
            $callable,
            WebapiException::class,
            fn (WebapiException $e) => assertThat($e->getHttpCode(), equalTo($httpStatusCode))
        );
    }

    private function assertThrowsException(callable $callable, string $exception, callable $exceptionCallable = null): void
    {
        try {
            $callable();
        } catch (Exception $e) {
            assertThat($e, isInstanceOf($exception));
            if ($exceptionCallable != null) {
                $exceptionCallable($e);
            }
        }
    }

    private function orderSyncResponse(string $status = 'IN_PROGRESS'): callable
    {
        return fn (string $syncId, DateTime $createdFrom, DateTime $createdTo, int $initialCount, string $clientCode): array => [
            'syncId' => $syncId,
            'createdFrom' => $createdFrom->format(DateTimeInterface::ATOM),
            'createdTo' => $createdTo->format(DateTimeInterface::ATOM),
            'initialCount' => $initialCount,
            'clientCode' => $clientCode,
            'processedCount' => 0,
            'errorCount' => 0,
            'status' => $status,
        ];
    }
}
